import { EditorState } from '@codemirror/state';
import { ViewUpdate } from '@codemirror/view';
import { EditorView } from 'codemirror';
import { html } from '@codemirror/lang-html';
import { glsl, getBaseExtensions } from '../util/codemirror';
import { type SupportedLanguage, type EditorConfig, type UniformTuple } from '../types';

export default class Editor {
  private language: SupportedLanguage;
  private doc: string;
  private uniforms: UniformTuple[];
  private state: EditorState;
  private onUpdate: (a: string) => unknown;
  private parent: HTMLElement;
  private view: EditorView;

  constructor(config: EditorConfig = { parent: '#app' }) {
    const parent: HTMLElement =
      typeof config.parent === 'string' ? document.querySelector(config.parent) || document.body : config.parent || document.body;

    const { doc, uniforms, onUpdate, language } = {
      doc: '',
      uniforms: [] as UniformTuple[],
      onUpdate: (s: string) => s,
      language: 'GLSL' as SupportedLanguage,
      ...config,
    };

    this.language = language as SupportedLanguage;
    this.onUpdate = onUpdate.bind(this);
    this.parent = parent;
    this.doc = doc || '';
    this.uniforms = uniforms;
    this.state = this.createState();
    this.view = new EditorView({
      parent: this.parent,
      state: this.state,
    });
  }

  getLanguage(): any {
    switch (this.language) {
      case 'GLSL':
        return glsl(this.uniforms);
      case 'HTML':
        return html();
    }
  }

  createState() {
    const extensions = [
      ...getBaseExtensions({ parent: this.parent }),
      this.getLanguage(),
      EditorView.updateListener.of(this.update.bind(this)),
    ];

    return EditorState.create({ doc: this.doc.trim(), extensions });
  }

  update(e: ViewUpdate) {
    if (typeof this.onUpdate !== 'function') return;
    const { state, docChanged } = e;
    if (!docChanged) return;
    this.onUpdate(state.doc.toString());
  }

  destroy() {
    this.view.destroy();
  }
}
